from __future__ import annotations

import joblib

import numpy as np
import scipy.stats
from scipy.signal import fftconvolve
from scipy.signal import find_peaks_cwt
from scipy.optimize import fmin_l_bfgs_b
from sklearn.decomposition import PCA, IncrementalPCA
from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.base import RegressorMixin
from sklearn.kernel_approximation import Nystroem
from sklearn.linear_model import ARDRegression
#from sklearn.svm import LinearSVR
#from sklearn.gaussian_process import GaussianProcessRegressor
#from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel
from itertools import product
from sklearn.model_selection import train_test_split

from sklearn.base import clone, MetaEstimatorMixin
from joblib import Parallel, delayed
from functools import partial
import multiprocessing as mp
from copy import deepcopy

from typing import Any, Dict, List, Optional, Union, Tuple

# previous method with a single matrix, but using gradient descent
# ARDRegression to be used instead
# advantages:
# * parallelization
# * using evidence maximization as an iterative method
# * automatic relevance determination to reduce overtraining
#
#from autograd import numpy as anp
#from autograd import grad
#class FitModel(RegressorMixin, BaseEstimator):
#    """
#    Linear regression model with uncertainties.
#
#    Args:
#      l: Regularization coefficient.
#    """
#    def __init__(self, l: float=1e-6):
#        self.l = l
#
#        # model parameter sizes
#        self.Nx: int = 0
#        self.Ny: int = 0
#
#        # fit result
#        self.A_inf: Optional[np.ndarray] = None
#        self.b_inf: Optional[np.ndarray] = None
#        self.u_inf: Optional[np.ndarray] = None
#
#        # fit monitoring
#        self.loss_train: List[float] = list()
#        self.loss_test: List[float] = list()
#
#        self.input_data = None
#
#    def fit(self, X: np.ndarray, y: np.ndarray, **fit_params) -> RegressorMixin:
#        """
#        Perform the fit and evaluate uncertainties with the test set.
#
#        Args:
#          X: The input.
#          y: The target.
#          fit_params: If it contains X_test and y_test, they are used to validate the model.
#
#        Returns: The object itself.
#        """
#        if 'X_test' in fit_params and 'y_test' in fit_params:
#            X_test = fit_params['X_test']
#            y_test = fit_params['y_test']
#        else:
#            X_test = X
#            y_test = y
#
#        # model parameter sizes
#        self.Nx: int = int(X.shape[1])
#        self.Ny: int = int(y.shape[1])
#
#        # initial parameter values
#        A0: np.ndarray = np.eye(self.Nx, self.Ny).reshape(self.Nx*self.Ny)
#        b0: np.ndarray = np.zeros(self.Ny)
#        Aeps: np.ndarray = np.zeros(self.Nx)
#        u0: np.ndarray = np.zeros(self.Ny)
#        x0: np.ndarray = np.concatenate((A0, b0, u0, Aeps), axis=0)
#
#        # reset loss monitoring
#        self.loss_train: List[float] = list()
#        self.loss_test: List[float] = list()
#
#        def loss(x: np.ndarray, X: np.ndarray, Y: np.ndarray) -> float:
#            """
#            Calculate the loss function value for a given parameter set `x`.
#
#            Args:
#              x: The parameters as a flat array.
#              X: The independent-variable dataset.
#              Y: The dependent-variable dataset.
#
#            Returns: The loss.
#            """
#            # diag( (in @ x - out) @ (in @ x - out)^T )
#            A = x[:self.Nx*self.Ny].reshape((self.Nx, self.Ny))
#            b = x[self.Nx*self.Ny:(self.Nx*self.Ny+self.Ny)].reshape((1, self.Ny))
#
#            b_eps = x[(self.Nx*self.Ny+self.Ny):(self.Nx*self.Ny+self.Ny+self.Ny)].reshape((1, self.Ny))
#            A_eps = x[(self.Nx*self.Ny+self.Ny+self.Ny):].reshape((self.Nx, 1))
#            log_unc = anp.matmul(X, A_eps) + b_eps
#
#            #log_unc = anp.log(anp.exp(log_unc) + anp.exp(log_eps))
#            iunc2 = anp.exp(-2*log_unc)
#
#            L = anp.mean( (0.5*((anp.matmul(X, A) + b - Y)**2)*iunc2 + log_unc).sum(axis=1), axis=0 )
#            weights2 = (anp.sum(anp.square(A.ravel()))
#                        #+ anp.sum(anp.square(b.ravel()))
#                        #+ anp.sum(anp.square(A_eps.ravel()))
#                        #+ anp.sum(anp.square(b_eps.ravel()))
#                        )
#            return L + self.l/2 * weights2
#
#        def loss_history(x: np.ndarray) -> float:
#            """
#            Calculate the loss function and keep a history of it in training and testing.
#
#            Args:
#              x: Parameters flattened out.
#
#            Returns: The loss value.
#            """
#            l_train = loss(x, X, y)
#            l_test = loss(x, X_test, y_test)
#
#            self.loss_train += [l_train]
#            self.loss_test += [l_test]
#            return l_train
#
#        def loss_train(x: np.ndarray) -> float:
#            """
#            Calculate the loss function for the training dataset.
#
#            Args:
#              x: Parameters flattened out.
#
#            Returns: The loss value.
#            """
#            l_train = loss(x, X, y)
#            return l_train
#
#        grad_loss = grad(loss_train)
#        sc_op = fmin_l_bfgs_b(loss_history,
#                              x0,
#                              grad_loss,
#                              disp=True,
#                              #factr=1e7,
#                              factr=10,
#                              maxiter=100,
#                              iprint=0)
#
#        # Inference
#        self.A_inf = sc_op[0][:self.Nx*self.Ny].reshape(self.Nx, self.Ny)
#        self.b_inf = sc_op[0][self.Nx*self.Ny:(self.Nx*self.Ny+self.Ny)].reshape(1, self.Ny)
#        self.u_inf = sc_op[0][(self.Nx*self.Ny+self.Ny):(self.Nx*self.Ny+self.Ny+self.Ny)].reshape(1, self.Ny) # removed np.exp
#        self.A_eps = sc_op[0][(self.Nx*self.Ny+self.Ny+self.Ny):].reshape(self.Nx, 1)
#
#    def predict(self, X: np.ndarray, return_std: bool=False) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
#        """
#        Predict y from X.
#
#        Args:
#          X: Input dataset.
#
#        Returns: Predicted Y and, if return_std is True, also its uncertainty.
#        """
#
#        # result
#        y = (X @ self.A_inf + self.b_inf)
#        if not return_std:
#            return y
#        # flat uncertainty
#        y_unc = self.u_inf[0,:]
#        # input-dependent uncertainty
#        y_eps = np.exp(X @ self.A_eps + y_unc)
#        return y, y_eps
#


def matching_ids(a: np.ndarray, b: np.ndarray, c: np.ndarray) -> np.ndarray:
    """Returns list of train IDs common to sets a, b and c."""
    unique_ids = list(set(a).intersection(b).intersection(c))
    return np.array(unique_ids)

def matching_two_ids(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """Returns list of train IDs common to sets a and b."""
    unique_ids = list(set(a).intersection(b))
    return np.array(unique_ids)

class PromptNotFoundError(Exception):
    """
    Exception representing the error condition generated by not finding the prompt peak.
    """
    def __init__(self):
        pass
    def __str__(self) -> str:
        return "No prompt peak has been detected."


class HighResolutionSmoother(TransformerMixin, BaseEstimator):
    """
    Smoothens out the high resolution data.

    Args:
      high_res_sigma: Energy resolution in eV.
    """
    def __init__(self,
                 high_res_sigma: float=0.2
                 ):
        self.high_res_sigma = high_res_sigma
        self.energy = None

    def fit(self, X, y=None, **fit_params) -> TransformerMixin:
        """
        Fit records the energy axis.

        Args:
          X: Irrelevant.
          y: Irrelevant.
          fit_params: Contains the energy axis in the key "energy" with shape (any, features).

        Returns: The object itself.
        """
        self.energy = np.copy(fit_params["energy"])
        if len(self.energy.shape) == 2:
            self.energy = self.energy[0,:]
        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        """
        Apply smoothing in X using the energy axis.

        Args:
          X: Input to smoothen with shape (train id, features).

        Returns: Smoothened out spectrum.
        """
        if self.high_res_sigma <= 0:
            return X
        # use a default energy axis is none is given
        # assume only the energy step
        energy = np.broadcast_to(self.energy, X.shape)

        # Apply smoothing
        n_features = X.shape[1]
        # get the centre value of the energy axis
        mu = energy[:, n_features//2, np.newaxis]
        # generate a gaussian
        gaussian = np.exp(-0.5*(energy - mu)**2/self.high_res_sigma**2)
        gaussian /= np.sum(gaussian, axis=1, keepdims=True)
        # apply it to the data
        high_res_gc = fftconvolve(X, gaussian, mode="same", axes=1)
        return high_res_gc

    def inverse_transform(self, Xt: np.ndarray) -> np.ndarray:
        """
        Dummy function to allow for an inverse transform.

        Args:
          Xt: Transformed variable.

        Returns: Xt.
        """
        return Xt

class UncertaintyHolder(TransformerMixin, BaseEstimator):
    """
    Keep track of uncertainties.

    """
    def __init__(self):
        self.unc: np.ndarray = np.zeros((1, 0), dtype=float)

    def set_uncertainty(self, unc: np.ndarray):
        """
        Set the uncertainty.

        Args:
          unc: The uncertainty.
        """
        self.unc = np.copy(unc)

    def fit(self, X, y=None) -> TransformerMixin:
        """
        Does nothing.

        Args:
          X: Irrelevant.
          y: Irrelevant.

        Returns: Itself.
        """
        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        """
        Identity map.

        Args:
          X: The input.
        """
        return X

    def inverse_transform(self, X: np.ndarray) -> np.ndarray:
        """
        Identity map.

        Args:
          X: The input.
        """
        return X

    def uncertainty(self):
        """The uncertainty recorded."""
        return self.unc
    
class DataHolder(TransformerMixin, BaseEstimator):
    """
    Keep track of relevance dictionaries.

    """
    def __init__(self, data: Dict[str, Any]=dict()):
        self.data: Dict[str, Any] = data

    def set_data(self, data: Dict[str, Any]):
        """
        Set.

        Args:
          data: Data.
        """
        self.data = deepcopy(data)
        
    def get_data(self) -> Dict[str, Any]:
        """
        Get.

        """
        return self.data

    def fit(self, X, y=None) -> TransformerMixin:
        """
        Does nothing.

        Args:
          X: Irrelevant.
          y: Irrelevant.

        Returns: Itself.
        """
        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        """
        Identity map.

        Args:
          X: The input.
        """
        return X

    def inverse_transform(self, X: np.ndarray) -> np.ndarray:
        """
        Identity map.

        Args:
          X: The input.
        """
        return X



class SelectRelevantLowResolution(TransformerMixin, BaseEstimator):
    """
    Select only relevant entries in the low-resolution data.

    Args:
      channels: Selected channels to use as an input for the low resolution data.
      tof_start: Start looking at this index from the low-resolution spectrometer data.
                 Set to None to perform no selection
      delta_tof: Number of components to take from the low-resolution spectrometer.
                 Set to None to perform no selection.
    """
    def __init__(self,
                 channels:List[str]=[f"channel_{j}_{k}"
                                     for j, k in product(range(1, 5), ["A", "B", "C", "D"])],
                 tof_start: Optional[int]=None,
                 delta_tof: Optional[int]=300,
                 ):
        self.channels = channels
        self.tof_start = tof_start
        self.delta_tof = delta_tof
        self.mean = dict()
        self.std = dict()

    def transform(self, X: Dict[str, np.ndarray], keep_dictionary_structure: bool=False) -> np.ndarray:
        """
        Get a dictionary with the channel names for the inut low resolution data and output
        only the relevant input data in an array.

        Args:
          X: Dictionary with keys named channel_{i}_{k},
             where i is a number between 1 and 4 and k is a letter between A and D.
          keep_dictionary_structure: Whether to concatenate all channels, or keep them as a dictionary.

        Returns: Concatenated and pre-processed low-resolution data of shape (train_id, features).
        """
        if self.tof_start is None:
            raise NotImplementedError("The low-resolution data cannot be transformed before the prompt has been identified. Call the fit function first.")
        y = X
        if self.delta_tof is not None:
            first = max(0, self.tof_start - self.delta_tof)
            last = min(X[self.channels[0]].shape[1], self.tof_start + self.delta_tof)
            y = {channel: item[:, first:last] for channel, item in X.items()}
        if not keep_dictionary_structure:
            return np.concatenate(list(y.values()), axis=1)
        return y

    def estimate_prompt_peak(self, X: Dict[str, np.ndarray]) -> int:
        """
        Estimate the prompt peak index.

        Args:
          X: Low resolution data with a dictionary containing the channel names.

        Returns: The index.
        """
        # reduce on channel and on train ID
        sum_low_res = - np.mean(sum(list(X.values())), axis=0)
        axis = np.arange(0.0, sum_low_res.shape[0], 1.0)
        #widths = np.arange(10, 50, step=5)
        #peak_idx = find_peaks_cwt(sum_low_res, widths)
        gaussian = np.exp(-0.5*(axis - sum_low_res.shape[0]//2)**2/20**2)
        gaussian /= np.sum(gaussian, axis=0, keepdims=True)
        # apply it to the data
        smoothened = fftconvolve(sum_low_res, gaussian, mode="same", axes=0)
        peak_idx = [np.argmax(smoothened)]
        if len(peak_idx) < 1:
            raise PromptNotFoundError()
        peak_idx = sorted(peak_idx, key=lambda k: np.fabs(sum_low_res[k]), reverse=True)
        best_guess = int(peak_idx[0])
        # look around this estimate for the maximum
        # this is probably not necessary
        #min_search = max(best_guess - 10, 0)
        #max_search = min(best_guess + 10, len(sum_low_res))
        #restricted_arr = sum_low_res[min_search:max_search]
        #improved_guess = min_search + int(np.argmax(restricted_arr))
        return best_guess

    def fit(self, X: Dict[str, np.ndarray], y: Optional[Any]=None) -> TransformerMixin:
        """
        Estimate the prompt peak index.

        Args:
          X: Low resolution data with a dictionary containing the channel names.
          y: Ignored.

        Returns: The object itself.
        """
        print("Finding peaks")
        self.tof_start = self.estimate_prompt_peak(X)
        X_tr = self.transform(X, keep_dictionary_structure=True)
        self.mean = {ch: np.mean(X_tr[ch], axis=0, keepdims=True)
                     for ch in X.keys()}
        self.std = {ch: np.std(X_tr[ch], axis=0, keepdims=True)
                    for ch in X.keys()}
        print("Found peaks")
        return self

    def debug_peak_finding(self, X: Dict[str, np.ndarray], filename: str):
        """
        Produce image to understand if the peak finding step worked well.

        Args:
          X: Low resolution data with a dictionary containing the channel names.
          filename: The file name where to save the plot.

        """
        sum_low_res = - np.mean(sum(list(X.values())), axis=0)
        peak_idx = self.estimate_prompt_peak(X)
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=(8, 16))
        ax = plt.gca()
        ax.plot(np.arange(peak_idx-300, peak_idx+300),
                sum_low_res[peak_idx-300:peak_idx+300],
                c="b",
                label="Data")
        ax.set(title="",
               xlabel="Photon Spectrometer channel",
               ylabel="Sum of all Photon Spectrometer channels")
        plt.axvline(peak_idx,
                linewidth=3,
                ls="--",
                color='r',
                label="Peak position")
        ax.legend()
        plt.savefig(filename)
        plt.close(fig)

def _fit_estimator(estimator, X: np.ndarray, y: np.ndarray):
    estimator = clone(estimator)
    estimator.fit(X, y)
    return estimator

class MultiOutputWithStd(MetaEstimatorMixin, BaseEstimator):

    def __init__(self, estimator, *, n_jobs=-1):
        self.estimator = estimator
        self.n_jobs = n_jobs

    def fit(self, X: np.ndarray, y: np.ndarray):
        """Fit the model to data, separately for each output variable.

        Args:
          X: {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.
          y: {array-like, sparse matrix} of shape (n_samples, n_outputs)
            Multi-output targets. An indicator matrix turns on multilabel
            estimation.

        Returns: self.
        """
        if y.ndim == 1:
            raise ValueError(
                "y must have at least two dimensions for "
                "multi-output regression but has only one."
            )

        self.estimators_ = Parallel(n_jobs=self.n_jobs)(
            delayed(_fit_estimator)(
                self.estimator, X, y[:, i]
            )
            for i in range(y.shape[1])
        )

        return self

    def predict(self, X: np.ndarray, return_std: bool=False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Predict multi-output variable using model for each target variable.

        Args:
          X: {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.

        Returns: {array-like, sparse matrix} of shape (n_samples, n_outputs)
            Multi-output targets predicted across multiple predictors.
            Note: Separate models are generated for each predictor.
        """
        y = Parallel(n_jobs=self.n_jobs)(
            delayed(e.predict)(X, return_std) for e in self.estimators_
            #delayed(e.predict)(X) for e in self.estimators_
        )
        if return_std:
            y, unc = zip(*y)
            return np.asarray(y).T, np.asarray(unc).T

        return np.asarray(y).T

class Model(TransformerMixin, BaseEstimator):
    """
    Object representing a previous fit of the model to be used to predict high-resolution
    spectrum from a low-resolution one.

    Args:
      channels: Selected channels to use as an input for the low resolution data.
      n_pca_lr: Number of low-resolution data PCA components.
      n_pca_hr: Number of high-resolution data PCA components.
      high_res_sigma: Resolution of the high-resolution spectrometer in electron-Volts.
      tof_start: Start looking at this index from the low-resolution spectrometer data.
                 Set to None to perform no selection
      delta_tof: Number of components to take from the low-resolution spectrometer.
                 Set to None to perform no selection.
      validation_size: Fraction (number between 0 and 1) of the data to take for
                       validation and systematic uncertainty estimate.
      n_nonlinear_kernel: Number of nonlinear kernel components added at the preprocessing stage
                       to obtain nonlinearities as an input and improve the prediction.

    """
    def __init__(self,
                 channels:List[str]=[f"channel_{j}_{k}"
                                     for j, k in product(range(1, 5), ["A", "B", "C", "D"])],
                 n_pca_lr: int=600,
                 n_pca_hr: int=40,
                 high_res_sigma: float=0.2,
                 tof_start: Optional[int]=None,
                 delta_tof: Optional[int]=300,
                 validation_size: float=0.05,
                 n_nonlinear_kernel: int=0):
        # models
        x_model_steps = list()
        x_model_steps += [('select', SelectRelevantLowResolution(channels, tof_start, delta_tof))]
        if n_nonlinear_kernel > 0:
            x_model_steps += [('fex', Pipeline([('prepca', PCA(n_pca_lr, whiten=True)),
                                                ('nystroem', Nystroem(n_components=n_nonlinear_kernel, kernel='rbf', gamma=None, n_jobs=8)),
                                                ]))]
        x_model_steps += [
                          ('pca', PCA(n_pca_lr, whiten=True)),
                          #('pca', TruncatedSVD(n_pca_lr)),
                          ('unc', UncertaintyHolder()),
                          ]
        self.x_model = Pipeline(x_model_steps)
        self.y_model = Pipeline([
                                ('smoothen', HighResolutionSmoother(high_res_sigma)),
                                ('pca', PCA(n_pca_hr, whiten=True)),
                                ('unc', UncertaintyHolder()),
                                ])
        #self.fit_model = FitModel()
        self.fit_model = MultiOutputWithStd(ARDRegression(n_iter=30, tol=1e-5, verbose=True))

        self.channel_mean = {ch: np.nan for ch in channels}
        
        #self.channel_relevance = {ch: np.nan for ch in channels}

        # size of the test subset
        self.validation_size = validation_size

    def get_channels(self) -> List[str]:
        """Get channels used in training."""
        return self.x_model.named_steps["select"].channels

    def get_energy_values(self) -> np.ndarray:
        """Get x-axis of high-resolution data."""
        return self.y_model.named_steps["smoothen"].energy

    def get_low_resolution_range(self) -> Tuple[int, int]:
        """Get bin number with first and last relevant bins in the low-resolution spectrum."""
        first = (self.x_model.named_steps['select'].tof_start - self.x_model.named_steps['select'].delta_tof)
        last = (self.x_model.named_steps['select'].tof_start + self.x_model.named_steps['select'].delta_tof)
        return first, last

    def debug_peak_finding(self, low_res_data: Dict[str, np.ndarray], filename: str):
        """
        Produce image to understand if the peak finding step worked well.

        Args:
          low_res_data: Low resolution data with a dictionary containing the channel names.
          filename: The file name where to save the plot.

        """
        self.x_model['select'].debug_peak_finding(low_res_data, filename)

    def preprocess_high_res(self, high_res_data: np.ndarray) -> np.ndarray:
        """
        Preprocess high-resolution data to remove high requency components.

        Args:
          high_res_data: The high-resolution data.

        Returns: Smoothened spectrum.
        """
        return self.y_model['smoothen'].transform(high_res_data)

    def fit(self, low_res_data: Dict[str, np.ndarray], high_res_data: np.ndarray, high_res_photon_energy: np.ndarray) -> np.ndarray:
        """
        Train the model.

        Args:
          low_res_data: Low resolution data as a dictionary with the key set to `channel_{i}_{k}`,
                        where i is a number between 1 and 4 and k is a letter between A and D.
                        For each dictionary entry, a numpy array is expected with shape
                        (train_id, ToF channel).
          high_res_data: Reference high resolution data with a one-to-one match to the
                         low resolution data in the train_id dimension. Shape (train_id, ToF channel).
          high_res_photon_energy: Photon energy axis for the high-resolution data.

        Returns: Smoothened high resolution spectrum.
        """
        print("Fitting PCA on low-resolution data.")
        x_t = self.x_model.fit_transform(low_res_data)
        print("Fitting PCA on high-resolution data.")
        y_t = self.y_model.fit_transform(high_res_data, smoothen__energy=high_res_photon_energy)
        #self.fit_model.set_params(fex__gamma=1.0/float(x_t.shape[0]))
        print("Fitting model.")
        self.fit_model.fit(x_t, y_t)

        # calculate the effect of the PCA
        print("Calculate PCA unc. on high-resolution data.")
        high_res = self.y_model['smoothen'].transform(high_res_data)
        high_pca = self.y_model['pca'].transform(high_res)
        high_pca_rec = self.y_model['pca'].inverse_transform(high_pca)
        high_pca_unc = np.sqrt(np.mean((high_res - high_pca_rec)**2, axis=0, keepdims=True))
        self.y_model['unc'].set_uncertainty(high_pca_unc)

        print("Calculate PCA unc. on low-resolution data.")
        low_res = self.x_model['select'].transform(low_res_data)
        pca_model = self.x_model['pca']
        if 'fex' in self.x_model.named_steps:
            pca_model = self.x_model['fex'].named_steps['prepca']
        low_pca = pca_model.transform(low_res)
        low_pca_rec = pca_model.inverse_transform(low_pca)
        low_pca_unc =  np.mean(np.sqrt(np.mean((low_res - low_pca_rec)**2, axis=1, keepdims=True)), axis=0, keepdims=True)
        self.x_model['unc'].set_uncertainty(low_pca_unc)

        # for consistency check
        self.x_pca_mean = np.mean(low_pca, axis=0, keepdims=True)
        self.x_pca_std = np.std(low_pca, axis=0, keepdims=True)
        
        # for consistency check per channel
        #selection_model = self.x_model['select']
        #for channel in self.get_channels():
        #    self.channel_mean[channel] = np.mean(low_res_data[channel], axis=0, keepdims=True)
        #    print(f"Calculate PCA relevance on {channel}")
        #    # freeze input data in one channel only
        #    low_res_data_frozen = {ch: low_res_data[ch] if ch != channel
        #                               else np.repeat(self.channel_mean[channel], low_res_data[ch].shape[0], axis=0)
        #                           for ch in self.get_channels()}
        #    low_res = selection_model.transform(low_res_data_frozen)
        #    low_pca = pca_model.fit_transform(low_res)
        #    low_pca_rec = pca_model.inverse_transform(low_pca)
        #    low_pca_unc =  np.mean(np.sqrt(np.mean((low_res - low_pca_rec)**2, axis=1, keepdims=True)), axis=0, keepdims=True)
        #    self.channel_relevance[channel] = low_pca_unc
        print("End of fit.")

        return high_res

    #def get_channel_quality(self, channel: str, low_res_data: Dict[str, np.ndarray],
    #                        pca_model: PCA,
    #                        channel_relevance: Dict[str, float],
    #                        selection_model: SelectRelevantLowResolution,
    #                        channel_mean: Dict[str, np.ndarray]) -> float:
    #    """
    #    Get the compatibility for a single channel.
    #    Args:
    #      channel: The channel.
    #      low_res: The data in a dictionary.
    #      pca_model: The PCA model.
    #    Returns: the compatibility factor.
    #    """
    #    # freeze input data in one channel only
    #    low_res_data_frozen = {ch: low_res_data[ch] if ch != channel
    #                               else np.repeat(channel_mean[channel], low_res_data[ch].shape[0], axis=0)
    #                           for ch in low_res_data.keys()}
    #    low_res_selected = selection_model.transform(low_res_data_frozen)
    #    low_pca = pca_model.transform(low_res_selected)
    #    low_pca_rec = pca_model.inverse_transform(low_pca)
    #    low_pca_unc = channel_relevance[channel]
    #    low_pca_dev =  np.sqrt(np.mean((low_res_selected - low_pca_rec)**2, axis=1, keepdims=True))
    #    quality = low_pca_dev/low_pca_unc
    #    return quality

    def check_compatibility_per_channel(self, low_res_data: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        """
        Check if a new low-resolution data source is compatible with the one used in training, by
        comparing the effect of the trained PCA model on it, but do it per channel.

        Args:
          low_res_data: Low resolution data as in the fit step with shape (train_id, channel, ToF channel).

        Returns: Ratio of root-mean-squared-error of the data reconstruction using the existing PCA model and the one from the original model per channel.
        """
        selection_model = self.x_model['select']
        quality = {channel: 0.0 for channel in low_res_data.keys()}
        channels = list(low_res_data.keys())
        # check if each channel is close to the mean
        low_res_selected = selection_model.transform(low_res_data, keep_dictionary_structure=True)
        deviation = {ch: (low_res_selected[ch] - selection_model.mean[ch])/selection_model.std[ch]
                     for ch in channels}
        ndof = {ch: float(deviation[ch].shape[1] - 1)
                for ch in channels}
        chi2 = {ch: np.sum(deviation[ch]**2, axis=1, keepdims=True)
                for ch in channels}
        cutoff = {ch: scipy.stats.chi2.ppf(0.05, df=deviation[ch].shape[1] - 1)
                  for ch in channels}
        ts = {ch: (chi2[ch] - cutoff[ch])/ndof[ch]
              for ch in channels}
        return ts

        # checking channel relevance
        #pca_model = self.x_model['pca']
        #if 'fex' in self.x_model.named_steps:
        #    pca_model = self.x_model['fex'].named_steps['prepca']
        ##with mp.Pool(len(low_res_data.keys())) as p:
        #values = map(partial(self.get_channel_quality,
        #                     low_res_data=low_res_data,
        #                     pca_model=pca_model,
        #                     channel_relevance=self.channel_relevance,
        #                     selection_model=selection_model,
        #                     channel_mean=self.channel_mean
        #                    ), channels)
        #quality = dict(zip(channels, values))
        #return quality

    def check_compatibility(self, low_res_data: Dict[str, np.ndarray]) -> np.ndarray:
        """
        Check if a new low-resolution data source is compatible with the one used in training, by
        comparing the effect of the trained PCA model on it.

        Args:
          low_res_data: Low resolution data as in the fit step with shape (train_id, channel, ToF channel).

        Returns: Ratio of root-mean-squared-error of the data reconstruction using the existing PCA model and the one from the original model.
        """
        low_res = self.x_model['select'].transform(low_res_data)
        pca_model = self.x_model['pca']
        if 'fex' in self.x_model.named_steps:
            pca_model = self.x_model['fex'].named_steps['prepca']
        low_pca = pca_model.transform(low_res)
        
        deviation = (low_pca - self.x_pca_mean)/self.x_pca_std
        ndof = float(deviation.shape[1] - 1)
        chi2 = np.sum(deviation**2, axis=1, keepdims=True)
        cutoff = scipy.stats.chi2.ppf(0.05, df=deviation.shape[1] - 1)
        ts = (chi2 - cutoff)/ndof
        return ts
        
        #low_pca_rec = pca_model.inverse_transform(low_pca)
        #low_pca_unc = self.x_model['unc'].uncertainty()

        #fig = plt.figure(figsize=(8, 16))
        #ax = plt.gca()
        #ax.plot(low_res[0,...],
        #        c="b",
        #        label="LR")
        #ax.plot(low_pca_rec[0,...],
        #        c="r",
        #        label="LR rec.")
        #ax.set(title="",
        #       xlabel="Photon Spectrometer channel",
        #       ylabel="Low resolution spectrometer intensity")
        #ax.legend()
        #plt.savefig("check.png")
        #plt.close(fig)

        #low_pca_dev =  np.sqrt(np.mean((low_res - low_pca_rec)**2, axis=1, keepdims=True))
        #return low_pca_dev/low_pca_unc


    def predict(self, low_res_data: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        """
        Predict a high-resolution spectrum from a low resolution given one.
        The output includes the uncertainty in its second and third entries of the first dimension.

        Args:
          low_res_data: Low resolution data as in the fit step with shape (train_id, channel, ToF channel).

        Returns: High resolution data with shape (train_id, energy channel) in a dictionary containing
                 the expected prediction in key "expected", the stat. uncertainty in key "unc" and
                 a (1, energy channel) array for the PCA syst. uncertainty in key "pca".
        """
        low_pca = self.x_model.transform(low_res_data)
        high_pca, high_pca_unc = self.fit_model.predict(low_pca, return_std=True)
        #high_pca = self.fit_model.predict(low_pca)
        #high_pca_unc = 0
        n_trains = high_pca.shape[0]
        pca_y = np.concatenate((high_pca,
                                high_pca + high_pca_unc),
                               axis=0)
        high_res_predicted = self.y_model["pca"].inverse_transform(pca_y)
        expected = high_res_predicted[:n_trains, :]
        unc = high_res_predicted[n_trains:, :] - expected
        return dict(expected=expected,
                    unc=unc,
                    pca=self.y_model['unc'].uncertainty())

    def save(self, filename: str):
        """
        Save the fit model in a file.

        Args:
          filename: File name where to save this.
        """
        joblib.dump([self.x_model,
                     self.y_model,
                     self.fit_model,
                     DataHolder(self.channel_mean),
                     DataHolder(self.x_pca_mean),
                     DataHolder(self.x_pca_std),
                     #DataHolder(self.channel_relevance)
                     ], filename, compress='zlib')

    @staticmethod
    def load(filename: str) -> Model:
        """
        Load model from a file.

        Args:
          filename: Name of the file where to read the model from.

        Returns: A new model object.
        """
        x_model, y_model, fit_model, channel_mean, x_pca_mean, x_pca_std = joblib.load(filename)
        obj = Model()
        obj.x_model = x_model
        obj.y_model = y_model
        obj.fit_model = fit_model
        obj.channel_mean = channel_mean.get_data()
        obj.x_pca_mean = x_pca_mean.get_data()
        obj.x_pca_std = x_pca_std.get_data()
        #obj.channel_relevance = channel_relevance.get_data()
        return obj

